-- ダンスキャラクター・ステージ

-- ステージ一覧を取得
-- return Stage情報のテーブル
local function GetStageList(self)
    local function GetStage(targetPath)
        local returnList = {}
        local fileList = FILEMAN:GetDirListing(targetPath, true, false)
        for _,name in pairs(fileList) do
            local folder = string.format('%s%s/', targetPath, name)
            local file
            local overlayFile = nil
            local hasStage = false
            -- LoaderAが存在する場合はencisoシステム
            if FILEMAN:DoesFileExist(folder..'LoaderA.lua') then
                file = 'LoaderA.lua'
                hasStage = true
                -- キャラクターより上に描画するLoaderB（任意データ）
                if FILEMAN:DoesFileExist(folder..'LoaderB.lua') then
                    overlayFile = 'LoaderB.lua'
                end
                returnList[#returnList+1] = {Name = name, Type = 'enciso', Path = folder, File = file, Overlay = overlayFile}
            end
            -- model.txtがあればOutFoxシステム
            if not hasStage and FILEMAN:DoesFileExist(folder..'model.txt') then
                returnList[#returnList+1] = {Name = name, Type = 'outfox', Path = folder, File = 'model.txt', Overlay = nil}
                hasStage = true
            end
        end
        return returnList
    end
    local function HaveNotKeyInTable(tab, name)
        for _,v in pairs(tab) do
            if v.Name == name then
                return false
            end
        end
        return true
    end
    -- まずここにあるステージは全部取得
    local stageList = GetStage('/Appearance/DanceStages/')
    -- ここにあるステージでリストと同名のステージは無視する
    for _,v in pairs(GetStage('/DanceStages/')) do
        if HaveNotKeyInTable(stageList, v.Name) then
            stageList[#stageList+1] = v
        end
    end
    table.sort(stageList, function(a, b) return a.Name < b.Name end)
    return stageList
end

-- 指定した名前からステージ情報を取得
-- Offや存在しない場合はnil
local function NameToSgate(self, _name)
    --local name = GAMESTATE:IsDemonstration() and ((_name ~= 'Off') and 'Random' or 'Off') or _name
    local name = _name
    if name and name ~= 'Off' then
        local stageList = GetStageList(self)
        if name == 'Random' then
            return (#stageList > 0) and stageList[math.random(#stageList)] or nil
        end
        for _,v in pairs(stageList) do
            if v.Name == name then
                return v
            end
        end
    end
    return nil
end

-- 指定したIDからキャラクター情報を取得
local function IdToCharacter(self, id)
    local chara = (id == PLAYER_1 or id == PLAYER_2) and GAMESTATE:GetCharacter(id) or CHARMAN:GetCharacter(id)
    local charaType = W_DANCER:CharaType(chara)
    if charaType then
        return {
            Id    = id,
            Chara = chara,
            Type  = charaType,
        }
    end
    return nil
end

-- ステージのActor
-- p1: Stage名
-- return Actor
local function ActorStage(self, stageName, ...)
    local isOverlay = ...
    local stage = NameToSgate(self, stageName)
    if stage then
        if isOverlay and not stage.Overlay then
            return Def.Actor({})
        end
        if stage == 'Random' then
            local stageList = W_DANCER:StageList()
            stage = stageList[math.random(#stageList)]
        end
        if stage.Type == 'outfox' then
            return LoadActor(THEME:GetPathG('', 'ScreenGamePlay/dancer/stage/outfox'), stage.Path)
        elseif stage.Type == 'enciso' then
            return LoadActor(THEME:GetPathG('', 'ScreenGamePlay/dancer/stage/enciso'), isOverlay and stage.Path..stage.Overlay or stage.Path..stage.File)
        end
    end
    return Def.Actor({})
end

-- キャラクターのActor
-- p1: DancerID一覧（:区切り）
-- return Actor
local function ActorCharacter(self, dancerIds, stageType)
    local dancer = {}
    for _,v in pairs(split(':', dancerIds)) do
        dancer[#dancer+1] = IdToCharacter(self, v)
    end
    local width = ((stageType ~= '') and 20 or 20) + math.max(#dancer-2, 0)*((stageType ~= '') and 10 or 20) -- 人数が3人以上の時は少しキャラ配置横幅を増やす
    local actors = Def.ActorFrame({})
    local charaBpm = {}
    local timing = {}
    for i=1, #dancer do
        actors[#actors+1] = LoadActor(THEME:GetPathG('', 'ScreenGamePlay/dancer/character/'..dancer[i].Type), dancer[i].Chara, stageType)..{
            InitCommand = function(self)
                charaBpm[i] = -1
                self:x((#dancer > 1) and (width/2 - (width*(i-1)/(#dancer-1))) or 0)
            end,
            -- カメラ操作
            ChangedGameplaySongMessageCommand = function(self, params)
                timing[i] = params.Timing[dancer[i].Id] or nil
            end,
            -- 速度変更
            ScreenUpdateMessageCommand = function(self, params)
                local idIsPlayer = (dancer[i].Id == PLAYER_1 or dancer[i].Id == PLAYER_2)
                local bpm = (timing[i] and idIsPlayer)
                    and timing[i]:GetBPMAtBeat(params.PnBeat[dancer[i].Id])
                    or GAMESTATE:GetSongBPS()*60
                if (idIsPlayer and params.PnStop[dancer[i].Id]) or (not idIsPlayer and params.Stop) then
                    bpm = 0
                end
                if charaBpm[i] ~= bpm then
                    self:playcommand('ChangedBpm', {Bpm = bpm})
                    charaBpm[i] = bpm
                end
            end,
        }
    end
    return actors
end

-- キャラクターの種別を取得
-- p1: Character
-- return classic / enciso
local function GetCharacterType(self, character)
    if character then
        -- 3Dキャラかどうか(何故かEncisoシステムのNoneは落ちるので表示しないようにする)
        if character:GetModelPath() ~= '' and character:GetDisplayName() ~= 'None' and character:GetDisplayName() ~= 'default' then
            local charaType = 'classic'
            local f = RageFileUtil.CreateRageFile()
            f:Open(character:GetModelPath(), 1)
            -- Encisoシステム判定
            while(not f:AtEOF()) do
                if f:GetLine() == '"12.joint_HipMaster"' then
                    charaType = 'enciso'
                    break
                end
            end
            f:Close()
            f:destroy()
            return charaType
        end
    end
    return nil
end

return {
    Stage     = ActorStage,
    GetStage  = NameToSgate,
    StageList = GetStageList,
    Character = ActorCharacter,
    CharaType = GetCharacterType,
}

--[[
Dancer.lua

Copyright (c) 2021 A.C

This software is released under the MIT License.
https://opensource.org/licenses/mit-license.php
--]]
