-- 画像関連

-- 対象の拡張子一覧を作成
local function CreateTypeList(self, ...)
    local addList = ...
    addList = addList or {}
    local typeList = {'png', 'mp4', 'avi', 'mpg', 'mpeg', 'wmv', 'jpg', 'jpeg', 'gif', 'bmp'}
    for i=1, #typeList do
        addList[#addList+1] = typeList[i]
    end
    return addList
end

-- 楽曲からバナーを取得
local function GetBannerPathBySong(self, song)
    song = song or W:Song()
    local path = THEME:GetPathG('Common fallback', 'banner')
    if song and song:HasBanner() then
        path = song:GetBannerPath()
    end
    return path
end

-- 楽曲からジャケットを取得
local function GetJacketPathBySong(self, song, ...)
    local priorityBg = ...  -- BannerよりBGを優先
    if priorityBg == nil then
        priorityBg = true
    end
    song = song or W:Song()
    local path = THEME:GetPathG('Common fallback', 'jacket')
    if song then 
        path = song:HasJacket() and song:GetJacketPath() or nil
        if not path and priorityBg then
            path = song:HasBackground() and song:GetBackgroundPath() or nil
        end
        if not path then
            path = song:HasBanner()
                and song:GetBannerPath()
                or THEME:GetPathG('Common fallback', 'jacket.png')
        end
    end
    return path
end

-- コースからバナーを取得
local function GetBannerPathByCourse(self, course)
    course = course or GAMESTATE:GetCurrentCourse()
    local path = THEME:GetPathG('Common fallback', 'banner')
    if course then
        local fileName = string.match(course:GetCourseDir(), '(.*)%.crs')
        if W_OPTION:ReadPref('CourseBannerStrictCheck') and fileName then
            if FILEMAN:DoesFileExist(fileName..'.png') then
                path = fileName..'.png'
            elseif FILEMAN:DoesFileExist(fileName..'.mp4') then
                path = fileName..'.mp4'
            elseif course:HasBanner() then
                path = course:GetBannerPath()
            end
        elseif course:HasBanner() then
            path = course:GetBannerPath()
        end
    end
    return path
end

-- コースからバナーを取得
local function GetJacketPathByCourse(self, course)
    local path = nil
    if course then
        local fileName = string.match(course:GetCourseDir(), '(.*)%.crs')
        if W_OPTION:ReadPref('CourseBannerStrictCheck') and fileName then
            fileName = fileName..'-jacket'
            if FILEMAN:DoesFileExist(fileName..'.png') then
                path = fileName..'.png'
            elseif FILEMAN:DoesFileExist(fileName..'.mp4') then
                path = fileName..'.mp4'
            end
        end
    end
    return path or GetBannerPathByCourse(self, course)
end

-- 前方一致でファイルを取得
local function GetGraphicFiles(self, path, ...)
    local typeList = ...
    typeList = typeList or CreateTypeList(self)
    local dir, lowerFile = string.match(path, '(.+/)([^/]+)')
    if not dir or not lowerFile then
        return nil
    end
    lowerFile = string.lower(lowerFile)
    local fileList = FILEMAN:GetDirListing(string.match(path, '(.+/)[^/]+'))
    for i=1, #fileList do
        local lowerDirFile = string.lower(fileList[i])
        if string.find(lowerDirFile, lowerFile, 0, true) then
            for j=1, #typeList do
                if typeList[j] == '' and lowerFile == lowerDirFile then
                    return dir..fileList[i]
                elseif lowerFile..'.'..typeList[j] == lowerDirFile then
                    return dir..fileList[i]
                end
            end
        end
    end
    return nil
end

-- 比率を指定して画像描画
local function FixedRatioActor(self, w, h, ...)
    local path, cache = ...
    local width, height = w, h
    local function SetRatio(self)
        -- 期待縦横比
        local ratioE = 1.0 * width / height
        -- 実縦横比
        local ratioA = 1.0 * self:GetWidth() / self:GetHeight()
        if ratioA > ratioE then
            -- 実画像が横長（横をクリップ）
            local crop = 1.0 * height / self:GetHeight()
            self:zoomto(self:GetWidth() * crop, height)
            self:cropleft(crop/2)
            self:cropright(crop/2)
        elseif ratioA < ratioE then
            -- 期待縦横比が横長（縦をクリップ）
            local crop = 1.0 * width / self:GetWidth()
            self:zoomto(width, self:GetHeight() * crop)
            self:croptop(crop/2)
            self:cropbottom(crop/2)
        else
            -- 同一比率
            self:zoomto(width, height)
        end
    end
    return Def[cache and 'Banner' or 'Sprite']({
        OnCommand = function(self)
            if not path then
                return
            end
            self:LoadBackground(path)
            SetRatio(self)
        end,
        UpdateParamsCommand = function(self, params)
            if params.Width then
                width = params.Width
            end
            if params.Height then
                height = params.Height
            end
            if params.File then
                if cache then
                    self:LoadFromCachedBanner(params.File)
                else
                    self:LoadBackground(params.File)
                end
            end
            SetRatio(self)
        end,
    })
end

return {
    SongBanner   = GetBannerPathBySong,
    SongJacket   = GetJacketPathBySong,
    CourseBanner = GetBannerPathByCourse,
    CourseJacket = GetJacketPathByCourse,
    GetFile      = GetGraphicFiles,
    FixedRatio   = FixedRatioActor,
    TypeList     = CreateTypeList,
}

--[[
Graphics.lua

Copyright (c) 2021 A.C

This software is released under the MIT License.
https://opensource.org/licenses/mit-license.php
--]]
